
/*
 * Copyright (C) 1998,  The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Visualization
 *      Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University of California, Berkeley nor of the 
 *    Lawrence Berkeley National Laboratory may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *
 * This work is supported by the U. S. Department of Energy under contract 
 * number DE-AC03-76SF00098 between the U. S. Department of Energy and the 
 * University of California.
 *
 *	Author: Wes Bethel
 *		Lawrence Berkeley National Laboratory
 *              Berkeley, California 
 *
 *  "this software is 100% hand-crafted by a human being in the USA"
 *
 * This code is a modified version of the read_trackd.c code distributed
 * with the CAVE library.
 *
 * september 1998
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <limits.h>
#include <sys/prctl.h>
#include <sys/schedctl.h>
#include "trackd.h"

#include "cvdev.h"

static int tracker_shmid=-1, controller_shmid=-1;
static struct TRACKD_TRACKING	*tracker=NULL;
static struct TRACKD_CONTROLLER	*controller=NULL;


/*************************************************************************
 void trackd_cleanup()
	When the program is interrupted by SIGINT, this routine closes the
 connections and exits.
**************************************************************************/
void trackd_cleanup(void)
{
 if (tracker)
	if (shmdt(tracker) < 0)
		perror("Detaching 'tracker' shared memory");
 if (controller)
	if (shmdt(controller) < 0)
		perror("Detaching 'controller' shared memory");
 exit(-1);  
}

/*************************************************************************
 void get_tracker_mem()
	Attaches to shared memory for the sensors.  The global 'tracker'
  points to the chunk.
**************************************************************************/
void get_tracker_mem(CAVE_CONFIG_ST *config)
{
 tracker_shmid = shmget(config->TrackdKey, sizeof(struct TRACKD_TRACKING), 0);
 if (tracker_shmid < 0)
	{
	fprintf(stderr,"ERROR: can't get tracker shared memory\n");
	perror("shmget");
	return;
	}
 tracker = (struct TRACKD_TRACKING *) shmat(tracker_shmid,(void *) 0, 0);
 if (tracker == (struct TRACKD_TRACKING *) -1)
	{
	fprintf(stderr,"ERROR: can't attach tracker shared memory\n");
	perror("shmat");
	}
}

/*************************************************************************
 void get_controller_mem()
	Attaches to shared memory for the buttons/joystick.  The global
  'controller' points to the chunk.
**************************************************************************/
void get_controller_mem(CAVE_CONFIG_ST *config)
{
 controller_shmid = shmget(config->ControllerdKey, 
			sizeof(struct TRACKD_CONTROLLER), 0);
 if (controller_shmid < 0)
	{
	fprintf(stderr,"ERROR: can't get controller shared memory\n");
	perror("shmget");
	return;
	}
 controller = (struct TRACKD_CONTROLLER *) 
			shmat(controller_shmid,(void *) 0, 0);
 if (controller == (struct TRACKD_CONTROLLER *) -1)
	{
	fprintf(stderr,"ERROR: can't attach controller shared memory\n");
	perror("shmat");
	}
}

		

#if 0
main(int argc,char **argv)
{
 CAVE_CONFIG_ST config;

 if (!getenv("CAVEDEBUGCONFIG"))
        putenv("CAVEDEBUGCONFIG=OFF");
 CAVEConfigurationInit(&argc,argv,NULL,&config);
	
 get_tracker_mem(&config);
 get_controller_mem(&config);

 signal(SIGINT,trackd_cleanup); /* catch interrupts to close nicely */

 while (1)
	{
	int i,j;
	if (tracker)
		{
		printf("%d sensors\n",tracker->header.numSensors);
		for (j=0; j < tracker->header.numSensors; j++)
			printf("sensor %d: (%.3f,%.3f,%.3f) (%.3f,%.3f,%.3f)\n", j,
				tracker->sensor[j].x,
				tracker->sensor[j].y,
				tracker->sensor[j].z,
				tracker->sensor[j].elev,
				tracker->sensor[j].azim,
				tracker->sensor[j].roll);
		}
	if (controller)
		{

		printf("controller:");
		printf(" %d buttons, %d valuators: ",controller->controller.num_buttons,controller->controller.num_valuators);

		for (j=0; j < controller->controller.num_buttons; j++)
			fprintf(stderr," %d",controller->controller.button[j]);
		fprintf(stderr,"   ");
		for (j=0; j < controller->controller.num_valuators; j++)
		fprintf(stderr," %.3f",controller->controller.valuator[j]);
		fprintf(stderr,"\r");
		fflush(stderr);

		printf("\n");

		}
	/* sleep(1); */
	}
}
#endif


/*
 *
 */

static int loc_nbuttons;
static int loc_ntrackers;

int
open_cave_devices(int *ntrackers, int *nbuttons)
{
    CAVE_CONFIG_ST config;
    int fake_ac= 1;
    char fake_av[]={"fake_argv"};
    
    signal(SIGINT,trackd_cleanup); /* catch interrupts to close nicely */
    
    if (!getenv("CAVEDEBUGCONFIG"))
        putenv("CAVEDEBUGCONFIG=OFF");
    CAVEConfigurationInit(&fake_ac,fake_av,NULL,&config);
    
    get_tracker_mem(&config);
    get_controller_mem(&config);

    /* do a quick sanity check to make sure everything's ok */

    if (tracker == NULL || controller == NULL) /* didn't attach..gotta quit */
	return(-1);

    /* else we'll assume everything is ok */

    loc_ntrackers = *ntrackers = tracker->header.numSensors;
    loc_nbuttons = *nbuttons = controller->header.numButtons;
    

    return(0);

}

#define HEAD_INDEX 0
#define HAND_INDEX 1

int
get_head_data(float head_position[3],
	      float head_orientation[3])
{
    head_position[0] = tracker->sensor[HEAD_INDEX].x;
    head_position[1] = tracker->sensor[HEAD_INDEX].y;
    head_position[2] = tracker->sensor[HEAD_INDEX].z;
    head_orientation[0] = tracker->sensor[HEAD_INDEX].elev;
    head_orientation[1] = tracker->sensor[HEAD_INDEX].azim;
    head_orientation[2] = tracker->sensor[HEAD_INDEX].roll;
}

int
get_hand_data(float hand_position[3],
	      float hand_orientation[3])
{
    hand_position[0] = tracker->sensor[HAND_INDEX].x;
    hand_position[1] = tracker->sensor[HAND_INDEX].y;
    hand_position[2] = tracker->sensor[HAND_INDEX].z;
    hand_orientation[0] = tracker->sensor[HAND_INDEX].elev;
    hand_orientation[1] = tracker->sensor[HAND_INDEX].azim;
    hand_orientation[2] = tracker->sensor[HAND_INDEX].roll;
}

int
get_valuator(float *xval, float *yval)
{
    *xval = controller->controller.valuator[0];
    *yval = controller->controller.valuator[1];
}

int
get_buttons(int *buttons)
{
    int i,stat=0;
    for (i=0;i<loc_nbuttons;i++)
    {
	if ((buttons[i] = controller->controller.button[i]) != 0)
	    stat = 1;
    }
    return(stat);
}

int
close_cave(void)
{
    if (tracker)
	if (shmdt(tracker) < 0)
	    perror("Detaching 'tracker' shared memory");
    if (controller)
	if (shmdt(controller) < 0)
	    perror("Detaching 'controller' shared memory");
    return(1);
}

void
cv_compute_orientation_matrix(float v[3], /* input rot: elev, az, roll */
			      matrix4x4 *rot)
{
    /*
     * given input azimuth, elevation and roll, compute a
     * rotation matrix. rotations are applied in order of az, elev, roll
     */
    double t,c,s,xr,yr,zr;
    extern double fabs(),cos(),sin();
    matrix4x4 rx,ry,rz;

    
    xr = DEGREES_TO_RADIANS(v[0]);
    yr = DEGREES_TO_RADIANS(v[1]);
    zr = DEGREES_TO_RADIANS(v[2]);

    identity_4x4(&rx);
    t = -xr;			/* assume device events are in radians  */
    c = cos(t);
    s = sin(t);
    rx.m[1][1] = rx.m[2][2] = c;
    rx.m[1][2] = s;
    rx.m[2][1] = -1. * s;
    
    identity_4x4(&ry);
    t = -yr;			/* assume device events are in radians */
    c = cos(t);
    s = sin(t);
    ry.m[0][0] = ry.m[2][2] = c;
    ry.m[0][2] = -1 * s;
    ry.m[2][0] = s;
    
    identity_4x4(&rz);
    t = -zr;			/* assume device events are in radians */
    c = cos(t);
    s = sin(t);
    rz.m[0][0] = rz.m[1][1] = c;
    rz.m[0][1] = s;
    rz.m[1][0] = -1. * s;
    
    identity_4x4(rot);

    mmul_4x4(rot,&ry);
    mmul_4x4(rot,&rx);
    mmul_4x4(rot,&rz);
}

#define FIX(a) (((a) > 1.0) ? 1.0 : ((a) < -1.0) ? -1.0 : (a))

void
compute_delta_angles(matrix4x4 *last,
		     matrix4x4 *this,
		     float v[3])
{
    matrix4x4 last_inverse,delta;

    invert_4x4matrix(last,&last_inverse);
    mmul_4x4_alt(this,&last_inverse,&delta);

    /*
     * the delta matrix represents the transformation from the rotation
     * transformation in "last" to that of "this."  now, we'll pick out
     * approximate values of delta x,y,z axis rotation to go from
     * last to this.
     */

    /*
     * let the delta matrix be represented by:
     *
     * |  a   b   c   0  |
     * |  d   e   f   0  |
     * |  g   h   i   0  |
     * |  0   0   0   1  |
     *
     * consider the basis of the untransformed coordinate system
     * given by the vectors [0,0,1], [0,1,0] and [0,0,1]. we assume
     * that the transformation induced by the delta matrix is some
     * small amount of rotation, since we're collecting orientation
     * info very quickly from a VR input device. using this assumption,
     * we can estimate the amount of change from last->this in
     * delta by doing dot products of delta with the untransformed
     * basis vectors.
     *
     * note that this is an approximation because we probably won't
     * be able to derive a rotation matrix from x-delta,y-delta,z-delta
     * that is exactly the delta matrix.
     */

    {
	float a,b,c,d,e,f,g,h,i;
	double xrot,yrot,zrot;
	double t1,t2;
	extern double acos();

	a = delta.m[0][0];
	b = delta.m[0][1];
	c = delta.m[0][2];

	d = delta.m[1][0];
	e = delta.m[1][1];
	f = delta.m[1][2];

	g = delta.m[2][0];
	h = delta.m[2][1];
	i = delta.m[2][2];
	
	/*
	 * z-axis rotation: the dot product of the x' axis of delta with
	 *   the [1,0,0] basis vector gives the cosine of the z-axis
	 *   rotation. the sign of rotation is given by the sign of 'd'.
	 *
	 * NOTE: we would really like a direct solution to this.
	 */
#if 0
	zrot = a * last->m[0][0] + d*last->m[1][0] + g*last->m[2][0];
#endif
	t1 = FIX(a);
	t2 = FIX(e);
	t1 = acos(t1);
	t2 = acos(t2);

	zrot = (t1+t2)*0.5;
	zrot *= (d < 0.) ? 1.0 : -1.0;

/*
	 * y-axis rotation: the dot product of the z' axis of delta with
	 *   the [0,0,1] basis vector gives the cosine of the y-axis
	 *   rotation. the sign of rotation is given by the sign of 'c'.
	 */
	t1 = FIX(a);
	t2 = FIX(i);
	t1 = acos(t1);
	t2 = acos(t2);

	yrot = (t1+t2)*0.5;
#if 0
	yrot = i;
	yrot = d + e + f;
	yrot *= 0.3333;
	yrot = c*last->m[0][2] + f*last->m[1][2] + i*last->m[2][2];
#endif
	yrot *= (c > 0.) ? 1.0 : -1.0;

	/*
	 * x-axis rotation: the dot product of the y' axis of delta with
	 *   the [0,1,0] basis vector gives the cosine of the x-axis
	 *   rotation. the sign of rotation is given by the sign of 'h'.
	 */
#if 0
	xrot = e;
	xrot = b*last->m[0][1] + e*last->m[1][1] + h*last->m[2][1];
#endif
	t1 = FIX(e);
	t2 = FIX(i);
	t1 = acos(t1);
	t2 = acos(t2);
	xrot = (t1+t2)*0.5;
	xrot *= (h > 0.) ? 1.0 : -1.0;

	v[0] = xrot;
	v[1] = yrot;
	v[2] = zrot;
	
	if (isnan(v[0]) || isnan(v[1]) || isnan(v[2]))
	{
	    int i;
	    fprintf(stderr," got me a NaN \n");
	    i = 1;
	}
    }
}

